import { combineLatest, isObservable, of, pipe, merge, Observable, observable, from, throwError, concat } from 'rxjs';
import { map, mergeScan, switchMap, withLatestFrom, sample, tap, ignoreElements, mapTo, share, distinctUntilChanged } from 'rxjs/operators';
import { Component, createElement } from 'react';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = Object.setPrototypeOf ||
    ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
    function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = Object.assign || function __assign(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
    }
    return t;
};

var combineSources = function () {
    var sources = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        sources[_i] = arguments[_i];
    }
    return combineLatest.apply(void 0, sources).pipe(map(function (values) { return values.reduce(function (a, c) { return (__assign({}, a, c)); }, {}); }));
};

var patchScan = pipe(mergeScan(function (state, update) {
    if (update instanceof Function) {
        var result = update(state);
        if (isObservable(result)) {
            return result.pipe(switchMap(function (foo) {
                if (foo instanceof Function)
                    return of(foo(state));
                return of(foo);
            }));
        }
        return of(result);
    }
    return of(__assign({}, state, update));
}));

var spreadMap = function (overrides) {
    if (overrides === void 0) { overrides = {}; }
    return map(function (value) { return (__assign({}, value, overrides)); });
};

var curry = function (fn) { return function () {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    return args.length < fn.length ? curry(fn.bind.apply(fn, [null].concat(args))) : fn.apply(void 0, args);
}; };

var scanPlans = curry(function (plans, source) {
    return merge.apply(void 0, [source].concat(Object.values(plans))).pipe(patchScan, spreadMap(plans));
});

var fromPlan = function (otherPlan, selector) {
    return pipe(withLatestFrom(otherPlan, function (_, value) { return value; }), map(selector));
};

var tapWhen = function (notifier, fn) {
    var pipe$$1 = [];
    for (var _i = 2; _i < arguments.length; _i++) {
        pipe$$1[_i - 2] = arguments[_i];
    }
    return function (source) {
        source
            .pipe.apply(source, [sample(notifier)].concat(pipe$$1, [tap(fn)])).subscribe();
        return source;
    };
};

var toPlan = function (otherPlan, selector) {
    if (selector === void 0) { selector = function (x) { return x; }; }
    return pipe(map(selector), tap(otherPlan), ignoreElements());
};

function plan() {
    var _a;
    var operators = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        operators[_i] = arguments[_i];
    }
    var next;
    var first = operators[0], rest = operators.slice(1);
    var ops = first
        ? first instanceof Function
            ? operators
            : [mapTo(first)].concat(rest)
        : operators;
    var o$ = (_a = new Observable(function (observer) {
        next = function () {
            var arg = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                arg[_i] = arguments[_i];
            }
            observer.next.apply(observer, arg);
        };
    })).pipe.apply(_a, ops.concat([share()]));
    var unsubscribe = o$.subscribe();
    next["unsubscribe"] = unsubscribe;
    next[observable] = function () { return o$; };
    return next;
}

var isNotPlan = function (x) { return isObservable(x) && !(x instanceof Function); };

var Stream = (function (_super) {
    __extends(Stream, _super);
    function Stream(props, context, config) {
        var _this = _super.call(this, props, context) || this;
        _this._isMounted = false;
        var _a = config ? config : props, sourcePipe = _a.pipe, plans = _a.plans;
        var state$ = _this.configureSource(props, config).pipe(distinctUntilChanged(), plans ? scanPlans(plans) : function (x) { return x; }, sourcePipe || (function (x) { return x; }), map(function (state) { return (__assign({}, state, { children: state.children || state.render || props.children || props.render })); }));
        _this.subscription = state$.subscribe(function (state) {
            if (_this._isMounted) {
                _this.setState(function () { return state; });
            }
            else {
                _this.state = state;
            }
        });
        return _this;
    }
    Stream.prototype.configureSource = function (props, config) {
        var _a = (config
            ? config
            : props).source, source = _a === void 0 ? throwError("No source provided") : _a;
        return isNotPlan(source) ? source : from(source);
    };
    Stream.prototype.componentDidMount = function () {
        this._isMounted = true;
    };
    Stream.prototype.render = function () {
        return this.state ? createElement(this.state.children, this.state) : null;
    };
    Stream.prototype.componentWillUnmount = function () {
        if (this.subscription)
            this.subscription.unsubscribe();
    };
    return Stream;
}(Component));
var stream = function (source, pipe$$1, plans) { return function (props, context) {
    return new Stream(props, context, { source: source, pipe: pipe$$1, plans: plans });
}; };

var StreamProps = (function (_super) {
    __extends(StreamProps, _super);
    function StreamProps() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    StreamProps.prototype.configureSource = function (props) {
        this.updateProps = plan();
        return concat(of(props), this.updateProps);
    };
    StreamProps.prototype.componentDidUpdate = function () {
        this.updateProps(this.props);
    };
    return StreamProps;
}(Stream));
var streamProps = function (pipe$$1, plans) { return function (props, context) {
    return new StreamProps(props, context, { pipe: pipe$$1, plans: plans });
}; };

export { combineSources, scanPlans, fromPlan, patchScan, spreadMap, tapWhen, toPlan, plan, Stream, stream, StreamProps, streamProps };
